import os
import subprocess
import sys
from tempfile import NamedTemporaryFile

from flask import safe_join

from . import storage
from .error import ActionError, AdfotgError


ADF_SIZE = 901120

MAX_FFS_FILENAME = 30


class FileUploadOp:
    def __init__(self, source, pos=None, rename=None):
        self._source = source
        self._pos = pos
        self._rename = rename

        self._tempfile = None

    @classmethod
    def interpret_api(cls, base_dir, api_arg):
        if isinstance(api_arg, dict):
            op = cls(
                safe_join(base_dir, api_arg['name']),
                pos=cls._interpret_pos(api_arg),
                rename=api_arg.get('rename')
            )
        elif isinstance(api_arg, str):
            op = cls(safe_join(base_dir, api_arg))
        else:
            raise ActionError("unknown API argument for file upload op")
        op.validate()
        return op

    @staticmethod
    def _interpret_pos(api_arg):
        start = _api_int(api_arg.get('start'))
        length = _api_int(api_arg.get('length'))
        if start is None and length is None:
            return None
        else:
            return start, length

    def open(self):
        if self._pos is not None:
            self._tempfile = NamedTemporaryFile()
            start, length = self._pos
            with open(self._source, 'rb') as fsrc:
                if start is not None:
                    fsrc.seek(start)
                if length is not None:
                    data = fsrc.read(length)
                else:
                    data = fsrc.read()
                if not data:
                    raise AdfotgError("read less data than requested "
                                      "from '{}'".format(self._source))
                self._tempfile.write(data)
                self._tempfile.flush()

    def close(self):
        if self._tempfile is not None:
            self._tempfile.close()
            self._tempfile = None

    def command(self):
        cmd = [
            'write',
            self._sys_name,
            self._adf_name
        ]
        return cmd

    def validate(self):
        if len(self._adf_name) > MAX_FFS_FILENAME:
            raise ActionError(
                "Amiga filenames must be in {} character "
                "limit; '{}' is {}".format(
                    MAX_FFS_FILENAME, self._adf_name,
                    len(self._adf_name)))

    @property
    def _sys_name(self):
        if self._tempfile is not None:
            return self._tempfile.name
        else:
            return self._source

    @property
    def _adf_name(self):
        if self._rename is not None:
            return self._rename
        else:
            return os.path.basename(self._source)


def is_adf_path(filepath):
    # Is there a better way to do this?
    return (os.path.isfile(filepath) and
            os.path.getsize(filepath) and
            filepath.lower().endswith(".adf"))


def create_adf(adf_path, label, file_ops):
    if not adf_path.lower().endswith(".adf"):
        raise ValueError("ADF filename must have .adf extension")
    workdir = os.path.dirname(adf_path)
    adf_name = os.path.basename(adf_path)
    cmd_base = [
        'xdftool', adf_name,
        'create', '+',
        'format', label, 'ffs',
    ]
    file_commands = []
    try:
        for file_op in file_ops:
            # TODO Creating this in the caller but opening it here
            # breaks the RAII rule.
            file_op.open()
            file_commands.append('+')
            file_commands += file_op.command()
        print("calling command", cmd_base + file_commands, file=sys.stderr)
        env = dict(os.environ)
        env['PYTHONIOENCODING'] = 'utf-8'
        p = subprocess.Popen(
            cmd_base + file_commands,
            cwd=workdir, stdin=subprocess.DEVNULL, env=env,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, _ = p.communicate()
        exitcode = p.wait()
        if exitcode != 0:
            storage.unlink(adf_path, optional=True)
            raise AdfotgError(stdout)
    finally:
        for file_op in file_ops:
            file_op.close()


def _api_int(i):
    return int(i) if i is not None else None



class Interpretable:
    '''Mixin for Enums, interprets value either directly
    as the Enum instance or as Enum.value.
    '''
    @classmethod
    def interpret(cls, o):
        if o in cls:
            return o
        else:
            return cls(o)



from .util import Interpretable

from collections import namedtuple
from enum import Enum
import fnmatch
import os
import re


class FileEntryField(Interpretable, Enum):
    NAME = "name"
    SIZE = "size"
    MTIME = "mtime"

    @classmethod
    def dictify(cls, name, size, mtime):
        return {
            cls.NAME.value: name,
            cls.SIZE.value: size,
            cls.MTIME.value: mtime
        }


class Direction(Interpretable, Enum):
    ASCENDING = "asc"
    DESCENDING = "desc"

    def is_ascending(self):
        return self == self.ASCENDING


def listdir(dirpath, name_filter=None, sort=None):
    '''
    sort -- tuple of (FileEntryField, Direction) enums or None
    '''
    if name_filter is None:
        name_filter = _true
    files = [entry for entry in os.scandir(dirpath) if entry.is_file()]
    files = [
        _fileentry_to_dict(entry)
        for entry in files
        if name_filter(entry.name)
    ]
    if sort:
        field, direction = sort
        field = FileEntryField.interpret(field)
        direction = Direction.interpret(direction)
        files = list(sorted(files,
                            key=lambda f: _case_insensitive(f[field.value]),
                            reverse=not direction.is_ascending()))
    return files


def find(dirpath, pattern, case_sensitive=True):
    '''
    dirpath -- dir to scan, it's not recursive
    pattern -- glob pattern
    case_sensitive -- whether pattern match is case sensitive
    '''
    re_flags = 0 if case_sensitive else re.IGNORECASE
    pattern = re.compile(fnmatch.translate(pattern), re_flags)
    matches = []
    for entry in os.scandir(dirpath):
        if entry.is_file() and pattern.match(entry.name):
            matches.append(entry.name)
    return matches


def unlink(filepath, optional=False):
    try:
        os.unlink(filepath)
    except FileNotFoundError:
        if not optional:
            raise


def mount_point(path):
    if path:
        path = os.path.realpath(path)
        while path and not os.path.ismount(path):
            path = os.path.dirname(path)
    return path


FsStats = namedtuple('StorageStats', ['name', 'total', 'avail'])


def fs_stats(path):
    stat = os.statvfs(path)
    return FsStats(path,
                   stat.f_bsize * stat.f_blocks,
                   stat.f_bsize * stat.f_bavail)


def _fileentry_to_dict(entry):
    stat = entry.stat()
    return FileEntryField.dictify(entry.name, stat.st_size, int(stat.st_mtime))


def _true(*args, **kwargs):
    return True


def _case_insensitive(what):
    return what.lower() if isinstance(what, str) else what



class AdfotgError(Exception):
    '''Error base for all errors generated by this app.'''
    pass


class ActionError(AdfotgError):
    '''This error denotes an invalid operation caused by the user.

    The API server translates it to code 400.

    '''
    pass



# coding: utf-8
from . import config, version
from .os_install import install, InstallError
import adfotg

from optparse import OptionParser
import os
import sys

PROFILE_ENV = "{}_PROFILE".format(version.SHORTNAME)


def main():
    profile_file = os.environ.get(PROFILE_ENV)
    if profile_file:
        # If profiling is enabled, start it.
        import yappi
        yappi.start()
    try:
        print_version()
        # Parse config.
        options = parse_args()
        if options.version:
            exit(0)
        print("", file=sys.stderr)
        if options.install:
            try:
                install()
            except InstallError as e:
                print("Install failed: {}".format(e), file=sys.stderr)
            exit(0)
        try:
            config.load(options.cfgfile)
        except config.ConfigError as cfg_error:
            print(cfg_error, file=sys.stderr)
            exit(1)
        # Run app.
        port = options.port or config.config.port
        adfotg.app.run(host='0.0.0.0', port=port)
    finally:
        if profile_file:
            # If profiling was enabled, stop it and generate reports.
            yappi.stop()
            pstats = yappi.convert2pstats(yappi.get_func_stats())
            pstats.dump_stats(profile_file)
            with open(profile_file + ".threads", "w") as threads_file:
                yappi.get_thread_stats().print_all(threads_file)


def print_version(file=sys.stderr):
    print("{} ({}) {} ({})".format(
        version.FULLNAME, version.SHORTNAME,
        version.VERSION, version.YEARSPAN),
        file=file)
    print("On MIT License; no warranty", file=file)


def parse_args():
    opt_parser = OptionParser()
    opt_parser.add_option(
        '-c', '--cfg', dest='cfgfile',
        help=('path to config file; by default looks in: {}' .format(
            ', '.join(config.DEFAULT_CONFIG_LOCATIONS))))
    opt_parser.add_option(
        '-p', '--port', dest='port', type=int,
        help='port on which to host the service (overrides config)')
    opt_parser.add_option(
        '--install', dest='install', default=False, action='store_true',
        help='Integrate adfotg with the Raspbian Operating System')
    opt_parser.add_option(
        '-V', '--version', dest='version', default=False,
        action='store_true', help='display version and quit')
    options, _ = opt_parser.parse_args()
    return options



'''This application implements a functionality of converting a Raspberry
Pi Zero into a programmable USB drive with the intention of the primary
use in a Gotek floppy drive emulator in an Amiga computer.

Entire functionality is exposed over a RESTful API that can be accessed
through HTTP requests. Hereby are shortlisted all API endpoints with
their HTTP methods, params, possible responses and descriptions.

Generic errors are returned as Content-Type 'application/json' with
"error" field set to the error message. The error code can be 4xx or
500, depending on if the client made an incorrect request (error 4xx) or
the server encountered an invalid state that the client could not
predict or mitigate (error 500). Refer to the specific documentation
of each method for details about possible error codes, but expect that
error code 500 can be produced by any method.


== Definitions ==

-- Arguments --

- URL args are specified as a part of URL path
- Query args are specified in URL query
- Body args are specified as a JSON object sent as the request's body

-- FileEntry --

File listing returned by this API are returned in a FileEntry
object format.

  FileEntry {
    name: file name
    size: size in bytes, integer
    mtime: last modification time in seconds since the 1970 epoch
  }

-- Zones --

- Upload zone - where the generic files are stored upon upload
- ADF library - where the ADFs are stored
- Mount images zone - where the USB drive images are stored

-- Mount Image --

A mount image is a file that serves as a FAT32 formatted container.
It can be mounted on a directory in a Linux filesystem, manipulated
with mtools and, most importantly, mounted as a mass storage device
through USB On-The-Go using the g_mass_storage Kernel module.

Simply put: A mount image can be considered to be an USB stick image.


== Disclaimer ==

The format of this documentation doesn't adhere to any standard.

Please note that the API is still a WORK IN PROGRESS which means
it's **unstable**, may change at whim and the documentation may
be incomplete.


== Static Files ==

This application doesn't only provide a REST API service but is also
a HTTP server that may serve HTML or other content to be rendered
in the browser. The default 'site' implementation comes with several
files that are served as-is.

The application handles this by first trying to match the request
URL to an API endpoint. If no API endpoint matches, it will fall back
to a generic URL handler which tries to see if there's a static file
or a web page that matches the requested path. If yes, then this
static file or web page is served. If there's no such fall-back,
then 404 is returned.
'''
from . import app
from . import adf, mountimg, selfcheck, storage, version
from .config import config
from .error import AdfotgError, ActionError
from .mountimg import Mount, MountImage

from flask import jsonify, request, safe_join, send_file, send_from_directory

import os
import shutil
import tempfile
import traceback
import weakref


@app.route("/upload", methods=['POST'])
def upload():
    '''Upload a file to the upload zone or the ADF library.

    This is a "smart" function. It accepts file uploads and tries to
    organize them. It expects files to be uploaded using the HTTP
    standard "multipart/form-data" format. The uploaded files are
    analyzed to see if they can be identified as ADFs. ADFs are then
    stored in the ADF library, while all other files are moved to the
    upload zone.

    '''
    os.makedirs(config.upload_dir, exist_ok=True)
    for filename, file in request.files.items():
        dest_path = safe_join(config.upload_dir, filename)
        file.save(dest_path)
        if adf.is_adf_path(dest_path):
            os.makedirs(config.adf_dir, exist_ok=True)
            shutil.move(dest_path, safe_join(config.adf_dir, filename))
    return ""


@app.route("/upload", methods=['GET'])
def list_uploads():
    '''Gets a list of files in the upload zone.

    Query args (all optional):
    - sort -- sort field, valid values: name, size, mtime; defaults to name
    - dir -- sort direction, valid values: asc, desc; defaults to asc

    Returns: a list of FileEntry objects.
    '''
    sorting = _sorting()
    return jsonify(storage.listdir(config.upload_dir, sort=sorting))


@app.route("/upload", methods=["DELETE"])
def del_uploads():
    '''Bulk delete of uploads.

    Body args:
    - names -- list of strings denoting upload names to delete.

    Returns: list of tuples which can be either: (200, filename, '')
    or (error_code, filename, error). There are as many elements
    in the returned list as there are names in the request.
    '''
    filenames = request.get_json().get('names', [])
    return jsonify(_del_files(config.upload_dir, filenames))


@app.route("/upload/<name>", methods=["GET"])
def get_upload(name):
    '''Retrieve a file from the upload zone.

    URL args:
    - name -- name of the file to retrieve

    Returns: The file.
    '''
    return send_from_directory(config.upload_dir, name)


@app.route("/upload/<name>", methods=["DELETE"])
def del_upload(name):
    '''Delete a file from the upload zone; returns nothing on success.

    URL args:
    - name -- name of the file to delete
    '''
    os.unlink(safe_join(config.upload_dir, name))


@app.route("/adf", methods=["GET"])
def list_adfs():
    '''Get a list of ADFs.

    Query args (all optional):
    - filter -- name filter, matched as "contains case-insensitive";
      defaults to nothing which disables the filter
    - sort -- sort field, valid values: name, size, mtime; defaults to name
    - dir -- sort direction, valid values: asc, desc; defaults to asc
    - start -- denotes index at which to start returning the values.
      Must be 0 or greater, but can go beyond the total amount.
    - limit -- maximum amount of elements to return; must be greater
      than 0 if specified.

    Returns: An object: {
        listing: [{...}, {...}, ...],
        total: integer
    }
    `listing` is a list of FileEntry objects. `total` is a total number
    of entries without the limitation which is useful in a query with a
    `limit` to calculate pages.

    '''
    # TODO - recurse into subdirectories or support more ADF dirs than one.
    # Users could potentially store hundreds of those and pagination is required.
    name_filter = _name_filter()
    sorting = _sorting()

    start, limit = _pagination()
    full_list = storage.listdir(config.adf_dir, name_filter=name_filter,
                                sort=sorting)
    return jsonify(
        listing=_limit(full_list, start, limit),
        total=len(full_list))


@app.route("/adf_std", methods=["GET"])
def list_standard_adfs():
    return jsonify(_list_standard_adfs())


def _list_standard_adfs():
    '''Returns a list of standard ADFs that may be used in specialized
    cases. These ADFs have known names and when placed on the USB drive
    will be used in special ways.

    The ADFs are returned only if they exist in the ADF library.

    Currently, the only recognized standard ADF is "SELECTOR.ADF" for
    Cortex firmware. Gotek with Cortex will not work without
    the selector.

    Returns: a list of ADF names with case adjusted to the actual
    case of the filename. If multiple files have the same but
    differently cased filename, then a single but undefined name
    is returned. If there are no matches, the returned list is empty.
    '''
    STANDARD_ADFS = ["selector.adf"]
    matches = []
    for adf_name in STANDARD_ADFS:
        found = storage.find(config.adf_dir, adf_name,
                             case_sensitive=False)
        if found:
            matches.append(found[0])
    return matches


@app.route("/adf", methods=["DELETE"])
def del_adfs():
    '''Bulk delete of ADFs.

    Body args:
    - names -- list of strings denoting upload names to delete.

    Returns: list of tuples which can be either: (200, filename, '')
    or (error_code, filename, error). There are as many elements
    in the returned list as there are names in the request.
    '''
    filenames = request.get_json().get('names', [])
    return jsonify(_del_files(config.adf_dir, filenames))


@app.route("/adf/<name>", methods=["GET"])
def get_adf(name):
    '''Retrieve an ADF.

    URL args:
    - name -- name of the ADF to retrieve

    Returns: The ADF file.
    '''
    return send_from_directory(config.adf_dir, name)


@app.route("/adf/<name>", methods=["DELETE"])
def del_adf(name):
    '''Delete an ADF from the ADF library; returns nothing on success.

    URL args:
    - name -- name of the ADF to delete
    '''
    os.unlink(safe_join(config.adf_dir, name))


@app.route("/adf/<name>", methods=["POST"])
def create_adf(name):
    '''Create ADF with contents from the upload zone.

    This API requires xdftool.

    There's a limitation of max. 30 characters on a filename
    and label length. If a source file exceeds this limitation,
    it's allowed to use the renaming mechanism so that it fits.

    URL args:
    - name -- name of the ADF file to create. Must not exist.

    Body args:
    - label -- disk label assigned to the ADF
    - contents -- list of elements pointing to the files
      in the upload zone.

    The `contents` list can hold values of different types.

    - A simple string type denotes a filename of a file from the upload
      zone. When specified, this file is copied whole on to the ADF.

    - A {name: string, start: int, length: int, rename: string} object
      denotes a split operation. The file pointed out by `name` will be
      cut into a slice as determined by the [`start`; `length`] range.
      It will then be renamed in accordance to `rename` string. This
      is useful to slice files larger than one ADF into multiple ADFs.
      `start` and `length` can be unspecified if you only wish to rename,
      and `rename` also can be unspecified in which case the behavior
      will be the same as if a simple string was sent.

    - An empty list will result in empty but formatted ADF.

    Returns: nothing if successful

    Errors:

    - 400
      - if ADF already exists
      - if disk label is not specified or exceeds the limit
        of 30 characters
      - if a filename exceeds the limit of 30 characters
      - if an invalid argument is specified for a file operation

    - 500 can be returned if xdftool fails, even if that
      failure was caused by the client preparing the operations
      incorrectly.

    '''
    target_path = safe_join(config.adf_dir, name)
    if os.path.exists(target_path):
        raise ActionError("ADF '{}' already exists".format(name))
    label = request.get_json().get("label")
    if not label:
        raise ActionError("must specify label")
    if len(label) > adf.MAX_FFS_FILENAME:
        raise ActionError("label length exceeds limit; max '{}', is '{}'".format(
            adf.MAX_FFS_FILENAME, len(label)))
    contents = request.get_json().get("contents", [])
    file_ops = [adf.FileUploadOp.interpret_api(config.upload_dir, piece)
                for piece in contents]
    adf.create_adf(target_path, label, file_ops)
    return ''


@app.route("/adf/<name>/quickmount", methods=["POST"])
def quickmount_adf(name):
    '''Just mount a specified, single ADF.

    This performs following steps in one go:

    1. Unmounts any mount image if already mounted.
    2. Creates a new temporary mount image. This image is
       created in the normal directory for the mount images
       but is always named .QUICKMOUNT.

       The contents of the image include all standard ADFs
       as listed by the /adf_std endpoint, and the selected ADF.

       2.1. Old temporary mount image is deleted, if there is any.
    3. Mounts this temporary mount image.

    Returns: nothing if successful

    Errors:
    - 404 -- if the ADF is not found
    '''
    adf_path = safe_join(config.adf_dir, name)
    if not os.path.isfile(adf_path):
        return _apierr(404, "ADF '{}' cannot be found".format(name))
    old_mount = Mount.current()
    if old_mount.state() == mountimg.MountStatus.Mounted:
        old_mount.unmount()
    img = _quickmount_image()
    if img.exists():
        img.delete()
    os.makedirs(os.path.dirname(img.imagefile), exist_ok=True)
    img.pack(
        [safe_join(config.adf_dir, n) for n in _list_standard_adfs()] +
        [adf_path])
    img_mount = Mount(img.imagefile)
    img_mount.mount()
    return ""


@app.route("/mount", methods=["GET"])
def get_mounted_flash_drive():
    '''Obtain mount state and info on the currently mounted image if any.

    There can be only one image mounted at a time.

    This function returns two pieces of information:

    1. Is there anything mounted or is there any error in
       regards to image mounting?

    2. If there is anything mounted, return details on it.

    Returns: an object {
      status: a MountStatus value as string
      file: string, optional; name of the image file if any is mounted
      listing: optional; list of FileEntry for files contained in the
          image
      error: string, optional; present only if there's a problem with
          obtaining the mount status or the information on the mount
          image
    }

    MountStatus is an enum that has following values:

    - mounted -- a valid image is mounted
    - unmounted -- nothing is currently mounted
    - no_image -- an image is mounted but it doesn't exist
      in the filesystem
    - bad_image -- if a predictable error occurs when trying
      to obtain the mount information; this may indicate
      that there's something wrong with the mounted image
    - other_image_mounted -- an image outside of adfotg is mounted

    '''
    mount = Mount.current()
    imagefile = None
    listing = []
    try:
        if mount.has_image():
            imagefile = mount.imagefile
            if not imagefile.startswith(config.mount_images_dir):
                return jsonify(status=mountimg.MountStatus.OtherImageMounted.value,
                               error="mounted image is unknown to the app")
            imagefile = imagefile[len(config.mount_images_dir):].lstrip("/")
            listing = mount.list()
    except AdfotgError as e:
        traceback.print_exc()
        return jsonify(status=mountimg.MountStatus.BadImage.value,
                       file=imagefile,
                       error=str(e))
    else:
        return jsonify(status=mount.state().value,
                       file=imagefile,
                       listing=listing)


@app.route("/mount/<imgname>", methods=["POST"])
def mount_flash_drive(imgname):
    '''Request to mount a mount image specified by the imgname.

    It is invalid to attempt to mount an image if another image
    is already mounted. `/unmount` first.

    URL args:
    - imgname -- name of the mount image file. It must exist
      in the mount images zone.

    Returns: nothing if successful

    Errors:
    - 404 -- if the mount image is not found
    - 500 -- can be returned if the mount image is invalid;
      this should not happen unless the image was modified
      externally
    '''
    imagefile = safe_join(config.mount_images_dir, imgname)
    mountimg = MountImage(imagefile)
    if not mountimg.exists():
        return _apierr(404, "image not found")
    if not mountimg.is_valid():
        return _apierr(500, "tried to mount an invalid mass storage image")
    mount = Mount(imagefile)
    mount.mount()
    return ""


@app.route("/unmount", methods=["POST"])
def unmount_flash_drive():
    '''Unmount the currently mounted image.

    Returns: nothing of importance on success.

    Errors:
    - 400 -- if unmount is requested but nothing is mounted.
    '''
    mount = Mount.current()
    if mount.state() is mountimg.MountStatus.Mounted:
        mount.unmount()
    else:
        return _apierr(400, "cannot unmount as nothing is mounted")
    return 'OK'


@app.route("/mount_image", methods=["GET"])
def list_mount_images():
    '''Gets a list of files in the mount images zone.

    Query args (all optional):
    - filter -- name filter, matched as "contains case-insensitive";
      defaults to nothing which disables the filter
    - sort -- sort field, valid values: name, size, mtime; defaults to name
    - dir -- sort direction, valid values: asc, desc; defaults to asc
    - start -- denotes index at which to start returning the values.
      Must be 0 or greater, but can go beyond the total amount.
    - limit -- maximum amount of elements to return; must be greater
      than 0 if specified.

    Returns: An object: {
        listing: [{...}, {...}, ...],
        total: integer
    }
    `listing` is a list of FileEntry objects. `total` is a total number
    of entries without the limitation which is useful in a query with a
    `limit` to calculate pages.

    '''
    name_filter = _name_filter()
    sorting = _sorting()
    start, limit = _pagination()
    if not os.path.exists(config.mount_images_dir):
        # App controls this directory so if it doesn't exist
        # it's not necessarilly an error.
        return jsonify([])
    full_list = storage.listdir(config.mount_images_dir,
                                name_filter=name_filter,
                                sort=sorting)
    return jsonify(
        listing=_limit(full_list, start, limit),
        total=len(full_list)
    )


@app.route("/mount_image/<imgname>", methods=["GET"])
def get_mount_image(imgname):
    '''Retrieve a mount image from the mount image zone.

    URL args:
    - imgname -- name of the mount image to retrieve

    Returns: The mount image.

    Errors:
    - 404 -- if the mount image is not found
    '''
    return send_from_directory(config.mount_images_dir, imgname)


@app.route("/mount_image/<imgname>/contents", methods=["GET"])
def get_mount_image_contents(imgname):
    '''List contents of this mount image as if it was a directory.

    URL args:
    - imgname -- name of the mount image to list.

    Returns: a list of FileEntry objects.

    Errors:
    - 404 -- if the mount image is not found
    '''
    img = MountImage(safe_join(config.mount_images_dir, imgname))
    if not img.exists():
        return _apierr(404, "image not found")
    return jsonify(img.list())


@app.route("/mount_image/<imgname>/contents/<filename>", methods=["GET"])
def get_file_from_mount_image(imgname, filename):
    '''
    Retrieve a file from inside of the mount image and
    send it back to the client.

    URL args:
    - imgname -- name of the image from which the file will be extracted.
    - filename -- name of the file inside the mount image to extract.

    Returns: The file is sent as a HTTP attachment.

    Errors:
    - 404 -- if the mount image is not found
    '''
    img = MountImage(safe_join(config.mount_images_dir, imgname))
    if not img.exists():
        return _apierr(404, "image not found")

    def cleanup():
        tmpf.close()
    tmpf = tempfile.NamedTemporaryFile()
    weakref.finalize(tmpf, cleanup)
    img.unpack_file(filename, tmpf.name)
    os.fsync(tmpf)
    tmpf.seek(0)
    size = os.path.getsize(tmpf.name)
    response = send_file(tmpf, as_attachment=True,
                         attachment_filename=filename)
    response.headers['Content-Length'] = size
    return response


@app.route("/mount_image", methods=["DELETE"])
def del_mount_images():
    '''Bulk delete of mount images.

    Body args:
    - names -- list of strings denoting image names to delete.

    Returns: list of tuples which can be either: (200, filename, '')
    or (error_code, filename, error). There are as many elements
    in the returned list as there are images in the request.
    '''
    filenames = request.get_json().get('names', [])
    deleted = []
    for filename in filenames:
        mountimg = MountImage(safe_join(config.mount_images_dir, filename))
        if not mountimg.exists():
            deleted.append((404, filename, "image not found"))
        try:
            mountimg.delete()
        except Exception as e:
            deleted.append((500, filename, str(e)))
    return jsonify(deleted)


@app.route("/mount_image/<imgname>", methods=["DELETE"])
def del_mount_image(imgname):
    '''Delete a mount image from the mount image zone.

    URL args:
    - imgname -- name of the mount image to delete

    Returns: nothing if successful

    Errors:
    - 404 -- if the mount image is not found
    '''
    mountimg = MountImage(safe_join(config.mount_images_dir, imgname))
    if not mountimg.exists():
        return _apierr(404, "image not found")
    mountimg.delete()
    return ""


@app.route("/mount_image/<imgname>/pack_adfs", methods=["PUT"])
def mount_pack_flash_drive_image(imgname):
    '''Creates a new mount image with specified ADFs.

    This function takes ADFs from the ADF library and packs
    them into a new mount image. The ADFs in the library remain
    untouched and their copies are packed into the mount image.

    The ADFs are copied into the mount image in order in which they are
    present on the 'adfs' list. This order is important as Gotek will
    use this order to assign the floppy index to each ADF (this may
    vary depending on your Gotek's firmware).

    Body args:
    - adfs -- list of ADFs to put into the image.
      Specify only the names here as they are returned
      by GET /adf.

    Errors:
    - 400 -- if no ADF is specified
          -- if any of the ADFs is not found in the library
          -- if mount image with specified name already exists
    '''
    args = request.get_json()
    adfs = args.get("adfs")
    if not adfs:
        return _apierr(400, "no ADFs specified")
    adfs_paths = [
        safe_join(config.adf_dir, adf)
        for adf in adfs
    ]
    for adf_, adf_path in zip(adfs, adfs_paths):
        if not os.path.isfile(adf_path):
            return _apierr(400, "ADF '{}' not found".format(adf_))
    os.makedirs(config.mount_images_dir, exist_ok=True)
    imagefile = safe_join(config.mount_images_dir, imgname)
    image = MountImage(imagefile)
    if image.exists():
        return _apierr(400, "image '{}' already exists".format(imgname))
    image.pack(adfs_paths)
    return ""


@app.route("/filesystem", methods=['GET'])
def get_free_space():
    '''Details of file-system roots for all file-systems used by the app.

    Returned is a list of file-systems that are used for storage by
    the application. Each entry on the list contains:

    object {
      name: string; mount point of the filesystem
      total: int; total space in bytes
      avail: int; space available for use in bytes
    }
    '''
    paths = [
        config.adf_dir,
        config.mount_images_dir,
        config.upload_dir
    ]
    mount_points = set([storage.mount_point(p) for p in paths])
    fs_stats = [storage.fs_stats(mp) for mp in mount_points]
    fs_stats = sorted(fs_stats, key=lambda ss: ss.name)
    return jsonify([{
        'name': fs.name,
        'total': fs.total,
        'avail': fs.avail
    } for fs in fs_stats])


@app.route("/check")
def self_check():
    '''Returns PASS/FAIL status for various features.

    Some features may be dependant on external tools and this API checks
    if those tools are available. The return value is an object with
    keys refering to a feature and values are either an empty string
    denoting that everything is okay or an error message.

    Returns: {
        rpi, g_mass_storage, xdftool, mtools
    }
    - rpi -- are we running on a Raspberry Pi device
    - g_mass_storage -- is this kernel module available
    - xdftool -- is xdftool available
    - mtools -- are mtools installed; several tools needed to
      manipulate the USB drive images are checked
    - storage -- do all workspace directories exists and are
      they writable
    '''
    return jsonify(**selfcheck.self_check())


@app.route("/version")
def get_version():
    '''adfotg version.

    Returns: an object {
      version: x.y.z string denoting the version;
          this doesn't contain the Git commit hash.
      yearspan: production years span as a string;
          this is human-readable and for display purposes.
    }
    '''
    return jsonify(
        version=version.VERSION,
        yearspan=version.YEARSPAN,
        lastyear=str(version.YEARS[-1])
    )


@app.route("/help")
def api_help():
    '''Returns the help in text/plain format.'''
    from .apidoc import spec
    return spec.to_text(), 200, {"Content-Type": "text/plain"}


def _del_files(dirpath, filenames):
    deleted = []
    for filename in filenames:
        try:
            storage.unlink(safe_join(dirpath, filename))
        except FileNotFoundError:
            deleted.append((404, filename, "file not found"))
        except Exception as e:
            deleted.append((500, filename, str(e)))
    return deleted


def _quickmount_image():
    return MountImage(safe_join(config.mount_images_dir, ".QUICKMOUNT"))


def _sorting():
    sort = request.args.get("sort", "name")
    direction = request.args.get("dir", "asc")
    return sort, direction


def _pagination():
    start = request.args.get("start")
    limit = request.args.get("limit")
    return _validate_start(start), _validate_limit(limit)


def _name_filter():
    pattern = request.args.get("filter")
    if pattern:
        pattern = pattern.strip().lower()
        if pattern:
            def name_filter(name):
                return pattern in name.lower()
            return name_filter
    return None


def _limit(list_, start, limit):
    start = _validate_start(start)
    limit = _validate_limit(limit)
    if start is not None:
        list_ = list_[start:]
    if limit is not None:
        list_ = list_[:limit]
    return list_


def _validate_start(start):
    if start is not None:
        try:
            start = int(start)
        except ValueError:
            raise ActionError("starting index must be a number")
        if start < 0:
            raise ActionError("starting index must be 0 or greater")
    return start


def _validate_limit(limit):
    if limit is not None:
        try:
            limit = int(limit)
        except ValueError:
            raise ActionError("limit must be a number")
        if limit <= 0:
            raise ActionError("limit must be greater than 0")
    return limit


def _apierr(code, message):
    return (
        jsonify({"error": message}),
        code,
        {'Content-Type': 'application/json'}
    )



from . import app, api, version

from flask import url_for

import urllib


class _Spec:
    def __init__(self):
        self._routes = self._get_routes()

    def to_text(self):
        lines = []
        for url in self._api_urls:
            for route in self._apis_for_url(url):
                lines.append("{} -- {}".format(url, ",".join(route.methods)))
                if route.doc:
                    lines.append(route.doc)
            lines.append("")
        doc = '\n\n'.join([_preamble(), api.__doc__, '== Endpoints =='])
        doc += '\n\n'
        doc += '\n'.join(lines)
        return doc

    def _get_routes(self):
        return [_Route(rule) for rule in app.url_map.iter_rules()]

    @property
    def _apis(self):
        return [r for r in self._routes if r.api_func is not None]

    @property
    def _api_urls(self):
        all_urls = set([r.url for r in self._apis])
        return sorted(list(all_urls))

    def _apis_for_url(self, url):
        def _sortkey(rule):
            methods = rule.methods
            known_methods = ['GET', 'DELETE', 'POST', 'PUT']
            for m in known_methods:
                if m in rule.methods:
                    return m
            return methods[0]
        return sorted([r for r in self._apis if r.url == url],
                      key=_sortkey)


class _Route:
    def __init__(self, rule):
        self.args = rule.arguments
        self.endpoint = rule.endpoint
        self.methods = [m for m in rule.methods
                        if m not in ['OPTIONS', 'HEAD']]
        self.url = urllib.parse.unquote(url_for(
            rule.endpoint, **self._format_args(self.args)))
        try:
            self.api_func = getattr(api, self.endpoint)
        except AttributeError:
            self.api_func = None
        if self.api_func:
            self.doc = self.api_func.__doc__

    def _format_args(self, args):
        return {arg: "{" + arg + "}" for arg in args}


def _preamble():
    return "{} version {} ({})".format(version.FULLNAME, version.VERSION, version.YEARSPAN)


spec = _Spec()



from .error import ActionError, AdfotgError
from .storage import FileEntryField

from enum import Enum
import os
import platform
import time
import shlex
import subprocess


class MountStatus(Enum):
    Mounted = "mounted"
    Unmounted = "unmounted"
    NoImage = "no_image"
    BadImage = "bad_image"
    OtherImageMounted = "other_image_mounted"


class Mount:
    @classmethod
    def current(cls):
        mounter = _mk_mounter()
        imagefile = mounter.mounted()
        return cls(imagefile)

    def __init__(self, mountfile):
        self._mounter = _mk_mounter()
        self._mountimage = MountImage(mountfile)

    @property
    def imagefile(self):
        return self._mountimage.imagefile

    def has_image(self):
        return self._mountimage.has_image()

    def state(self):
        mounted = self._mounter.mounted()
        if mounted == self.imagefile:
            if mounted is None:
                return MountStatus.Unmounted
            else:
                return MountStatus.Mounted
        elif not self._mountimage.exists():
            return MountStatus.NoImage
        elif mounted:
            return MountStatus.OtherImageMounted
        else:
            # Image exists but is not mounted.
            return MountStatus.Unmounted

    def mount(self):
        mount_state = self.state()
        if mount_state == MountStatus.Mounted:
            # Image is already mounted, no-op.
            return
        elif mount_state == MountStatus.OtherImageMounted:
            raise ActionError("an image is already mounted; unmount it first")
        elif mount_state == MountStatus.NoImage:
            raise ActionError("image doesn't exist, cannot mount")
        self._mounter.mount(self._mountimage.imagefile)

    def unmount(self):
        if self.state() != MountStatus.Mounted:
            raise ActionError("cannot unmount when not mounted")
        self._mounter.unmount()

    def list(self):
        if self._mountimage.exists():
            return self._mountimage.list()
        else:
            return []

    def delete_image(self):
        self._mountimage.delete()


class MountImage:
    '''Requires mtools, because 'mount' requires root and mtools don't.'''
    # TODO do we need that? How much do we need that?
    _BUFFER_SPACE = 1024 * 1024
    _SECTORS_PER_TRACK = 32
    _SECTOR_SIZE = 512
    _SECTOR_ALIGNMENT = _SECTORS_PER_TRACK * _SECTOR_SIZE

    def __init__(self, imagefile):
        self._imagefile = imagefile

    @property
    def imagefile(self):
        return self._imagefile

    def delete(self):
        try:
            os.unlink(self._imagefile)
        except FileNotFoundError:
            pass

    def exists(self):
        return os.path.isfile(self._imagefile)

    def has_image(self):
        return self._imagefile is not None

    def is_valid(self):
        try:
            self.list()
            return True
        except AdfotgError:
            return False

    def list(self):
        '''
        List contents of the image as FileEntry-compatible list
        of dicts.
        '''
        p = subprocess.Popen(['mdir', '-i', self._imagefile],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = p.communicate()
        if p.wait() != 0:
            raise AdfotgError("'mdir' ended with non-zero exit code: {}".format(
                stderr))
        stdout = stdout.decode('utf-8')
        return _MdirParser().parse(stdout)

    def pack(self, files):
        if not isinstance(files, list):
            raise ValueError("files argument must be a list")
        sum_size = 0
        for file in files:
            sum_size += os.path.getsize(file)
        sum_size -= sum_size % self._SECTOR_ALIGNMENT
        sum_size += self._BUFFER_SPACE
        self.delete()
        self._create(size=sum_size)
        for file in files:
            subprocess.check_call(['mcopy', '-i', self._imagefile,
                                   file, '::'])

    def unpack(self, destdir):
        files = self.list()
        for file in files:
            self._mcopy(file, destdir)

    def unpack_file(self, file, destfile):
        if os.path.isdir(destfile):
            raise AdfotgError("target path '{}' is a directory".format(
                destfile))
        self._mcopy(file, destfile)

    def _mcopy(self, file, dest):
        subprocess.check_call(['mcopy', '-i', self._imagefile, '-n',
                               '::/{}'.format(file), dest])

    def _create(self, size=64 * 1024 * 1024):
        with open(self._imagefile, 'wb') as f:
            f.write(b'\0' * size)
        subprocess.check_call(['mkdosfs', self._imagefile])


class _RealMount:
    '''https://gist.github.com/gbaman/50b6cca61dd1c3f88f41

    root privileges are needed to manipulate kernel modules

    Currently mounted file is available in
    /sys/module/g_mass_storage/parameters/file
    '''
    def mounted(self):
        try:
            with open('/sys/module/g_mass_storage/parameters/file', 'r') as f:
                return f.read()
        except FileNotFoundError:
            return None

    def mount(self, imagefile):
        self.unmount()
        # It must be run with 'shell=True' as I couldn't figure
        # out how to make paths with whitespace work otherwise.
        cmd = ('sudo modprobe g_mass_storage file={} '
               'stall=0 removable=1'.format(
                   shlex.quote(imagefile)))
        subprocess.check_call(cmd, shell=True)

    def unmount(self):
        subprocess.check_call(['sudo', 'modprobe', '-r', 'g_mass_storage'])


class _FakeMount:
    # Mount state is global, this is true on the real
    # system, and should be true in the faker as well.
    _state = {
        'mounted': None
    }

    def mounted(self):
        return self._state['mounted']

    def mount(self, imagefile):
        self._state['mounted'] = imagefile

    def unmount(self):
        self._state['mounted'] = None


class _MdirParser:
    '''
    Parses output like this one:

      Volume in drive : has no label
      Volume Serial Number is 6515-5815
      Directory for ::/

      5years1         901120 2018-12-02  20:30
      5years1  adf    901120 2018-12-02  20:30
      adfotg1  adf    901120 2018-12-02  20:30
      BARBAR~1 ADF    901120 2018-11-24  22:44  Barbarian Plus 6.adf
      EOBAGA~1 ADF    901120 2018-11-24  22:44  Eob AGA z ppa01.adf
      GENESIA1 ADF    901120 2018-11-24  22:44  Genesia1.adf
              3 files           2 703 360 bytes
                                1 024 000 bytes free

    and returns it as storage.FileEntryField compatible list of dicts.
    '''
    def parse(self, output):
        listing = []
        has_directory = False
        for line in output.split("\n"):
            if not line:
                continue
            if "::/" in line:
                # Detect line that says "Directory for ::/"
                has_directory = True
                continue
            if not has_directory:
                # Skip until directory listing begins.
                continue
            try:
                entry = self.parse_entry(line)
            except ValueError:
                # This may be a bad idea.
                pass
            else:
                if entry is not None:
                    listing.append(entry)
        return listing

    def parse_entry(self, line):
        '''
        Example entry:
          BARBAR~1 ADF    901120 2018-11-24  22:44  Barbarian Plus 6.adf

        Another example entry:
          5years1  adf    901120 2018-12-02  20:30

        Yet another example entry without an extension:
          5years1         901120 2018-12-02  20:30


        The entry without the long name gets listed when the filename
        fits in the 8.3 DOS limitation and doesn't have a whitespace
        character.
        '''
        DOS_NAME_LEN = 8
        DOS_EXT_LEN = 3
        DOS_LEN = DOS_NAME_LEN + 1 + DOS_EXT_LEN  # NAME DOT EXT
        dos_name = line[:DOS_LEN]

        line = line[DOS_LEN:]
        size_start_index = self._find_index_of_first_non_space(line, 0)
        if size_start_index < 0:
            raise ValueError("cannot parse line '{}' as file entry: "
                             "size not found".format(line))
        size_end_index = line.find(' ', size_start_index)
        size = int(line[size_start_index:size_end_index])
        line = line[size_end_index + 1:]  # +1 space
        # Products of a drunken stupor.
        date = line[:line.find('  ', line.find('  ') + 2)]
        # Programmers know how to party hard.
        name = line[len(date) + 2:]
        if not name:
            # We have the case where the name fit in the DOS limitations
            # and we need to parse dos_name.
            name = dos_name[:DOS_NAME_LEN].strip()
            ext = dos_name[DOS_NAME_LEN + 1:].rstrip()  # 1 is for the dot
            if ext:
                name += "." + ext
        timestamp = time.mktime(time.strptime(date, "%Y-%m-%d  %H:%M"))
        return FileEntryField.dictify(name, size, int(timestamp))

    def _find_index_of_first_non_space(self, haystack, start):
        for i in range(start, len(haystack)):
            if haystack[i] != ' ':
                return i
        return -1


def _mk_mounter():
    if _is_faking_it():
        return _FakeMount()
    else:
        return _RealMount()


def _is_faking_it():
    # On RPi Zero platform.machine() returns 'armv6l'.
    return ('arm' not in platform.machine() or
            'ADFOTG_FAKE' in os.environ)



import os
import pwd
import subprocess
import sys
from pkg_resources import resource_exists, resource_string
from subprocess import PIPE
from tempfile import mkdtemp

_BINARY_PATH = '/usr/local/bin/adfotg'
_CONF_RESOURCE = 'conf/adfotg.conf'
_CONF_DESTINATION = '/etc/adfotg.conf'
_HOME_DIR = '/var/lib/adfotg'
_OS_USER = 'adfotg'


class InstallError(Exception):
    pass


def install():
    _check_preconditions()
    _install()


def _check_preconditions():
    if not _is_root():
        raise InstallError("must run as root")
    if not _has_binary('adfotg'):
        raise InstallError("adfotg not found")
    os_id = _os_id()
    _log("OS ID:", os_id)
    if os_id != 'Raspbian':
        raise InstallError("OS must be Raspbian")
    if not os.path.isdir("/lib/systemd"):
        raise InstallError("systemd not detected")
    if not _has_res(_CONF_RESOURCE):
        _log("install package doesn't have the base config file; "
             "'{}' won't be created".format(_CONF_DESTINATION))
    if not os.path.exists(_BINARY_PATH):
        raise InstallError("adfotg executable not found at expected "
                           " path: {}".format(_BINARY_PATH))


def _install():
    _add_user()
    _allow_user_sudo()
    _install_conf()
    _install_service()
    _goodbye()


def _add_user():
    try:
        pwd.getpwnam(_OS_USER)
        return
    except KeyError:
        _log("OS user '{}' already exists; skipping".format(_OS_USER))
        pass
    _log("Adding OS user '{}'".format(_OS_USER))
    empty_dir = mkdtemp()
    try:
        cmd = [
            'useradd',
            '--system',
            '--skel', empty_dir,
            '--home-dir', _HOME_DIR,
            '-g', 'nogroup',
            '-G', 'sudo',
            '--create-home',
            _OS_USER
        ]
        _log('Creating OS user: {}'.format(' '.join(cmd)))
        subprocess.check_call(cmd)
    finally:
        os.rmdir(empty_dir)


def _allow_user_sudo():
    _log("Allowing passwordless sudo for OS user '{}'".format(_OS_USER))
    with open('/etc/sudoers.d/010_adfotg-nopasswd', 'w') as out:
        out.write("adfotg ALL=(ALL) NOPASSWD: ALL\n")


def _install_conf():
    if os.path.exists(_CONF_DESTINATION):
        _log("{} already exists; skipping".format(_CONF_DESTINATION))
        return
    _log("Creating config file {}".format(_CONF_DESTINATION))
    try:
        conf = _res(_CONF_RESOURCE)
    except FileNotFoundError:
        return
    with open(_CONF_DESTINATION, 'wb') as out:
        out.write(conf)


def _install_service():
    _log("Installing 'adfotg.service' in systemd")
    with open('/lib/systemd/system/adfotg.service', 'w') as out:
        out.write(_SYSTEMD_SERVICE)
    subprocess.check_call(['systemctl', 'enable', 'adfotg.service'])


def _goodbye():
    _log("")
    _log("Installation complete!")
    _log("It should be possible to start adfotg with 'sudo service adfotg start'")
    _log("Log is available through 'journalctl -fu adfotg'")


def _is_root():
    return os.geteuid() == 0


def _has_binary(name):
    return subprocess.call(['which', name]) == 0


def _has_user(name):
    pass


def _os_id():
    res = subprocess.run(['lsb_release', '-s', '-i'], stdout=PIPE,
                         universal_newlines=True)
    return res.stdout.strip()


def _has_res(name):
    return resource_exists('adfotg', name)


def _res(name):
    return resource_string('adfotg', name)


def _log(*args):
    print(*args, file=sys.stderr)


_SYSTEMD_SERVICE = """[Unit]
Description=ADF On-The-Go
After=network.target

[Service]
ExecStart={binary}
User={user}
SyslogIdentifier=adfotg
StandardOutput=syslog
StandardError=syslog

[Install]
WantedBy=multi-user.target
""".format(binary=_BINARY_PATH, user=_OS_USER)



'''
This module implements the self-check API.
'''
import os
import subprocess
import traceback

from .config import config


def self_check():
    return {
        'rpi': _check_rpi(),
        'g_mass_storage': _check_mass_storage(),
        'mtools': _check_mtools(),
        'xdftool': _check_xdftool(),
        'storage': _check_storage()
    }


def _catch_error(func):
    def _catcher(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            traceback.print_exc()
            return str(e)
    return _catcher


@_catch_error
def _check_rpi():
    '''Credits: https://raspberrypi.stackexchange.com/a/74541'''
    try:
        with open('/proc/cpuinfo', 'r') as cpuinfo:
            for line in cpuinfo:
                if line.startswith('Hardware'):
                    label, value = line.strip().split(':', 1)
                    value = value.strip()
                    if value not in ['BCM2708', 'BCM2709',
                                     'BCM2835', 'BCM2836']:
                        return (
                            'This system does not appear to be a '
                            'Raspberry Pi.'
                        )
                    else:
                        return ''
            return ('Could not find hardware information -- '
                    'this program may not be running on a Raspberry Pi.')
    except IOError as e:
        return 'Unable to open `/proc/cpuinfo`: {}'.format(e)
    return ''


@_catch_error
def _check_mass_storage():
    if 0 != _call(['modinfo', 'g_mass_storage']):
        return 'g_mass_storage Kernel module not found'
    return ''


@_catch_error
def _check_mtools():
    commands = ['mdir', 'mcopy', 'mkdosfs']
    return _check_commands(commands)


@_catch_error
def _check_xdftool():
    return _check_command('xdftool')


@_catch_error
def _check_storage():
    return _check_errors(_check_dir_writeable, config.all_workspace_dirs)


def _check_commands(cmds):
    return _check_errors(_check_command, cmds)


def _check_command(name):
    if 0 != _call(['which', name]):
        return '{} not found on $PATH'.format(name)
    return ''


def _check_dir_writeable(dirpath):
    if not os.path.isdir(dirpath):
        return "'{}' does not exist or is not a directory".format(dirpath)
    if not os.access(dirpath, os.W_OK):
        return "'{}' directory exists but is not writeable".format(dirpath)
    if not os.access(dirpath, os.R_OK):
        return "'{}' directory exists but is not readable".format(dirpath)
    return ''


def _check_errors(checker, elems):
    return '\n'.join(
        [error for error in map(checker, elems) if error])


def _call(command):
    devnull = subprocess.DEVNULL
    return subprocess.call(
        command,
        stdout=devnull,
        stderr=devnull,
        stdin=devnull)



#!/usr/bin/env python
# coding: utf-8
from setuptools import setup
import subprocess
import os


def pipen(cmd):
    # It's a popen, but with pipes.
    return subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)


def out_to_str(out):
    return out.decode('utf-8', 'replace').strip()


def get_git_tag():
    git = pipen(['git', 'describe', '--abbrev=0'])
    stdout, stderr = git.communicate()
    exitcode = git.wait()
    if exitcode == 128:
        return None
    elif exitcode != 0:
        raise Exception('fail: {}'.format(out_to_str(stderr)))
    else:
        return out_to_str(stdout)


def get_git_describe(tag):
    '''Returns the describe part without the tag.'''
    git = pipen(['git', 'describe', '--always', '--long', '--dirty'])
    stdout, stderr = git.communicate()
    exitcode = git.wait()
    if exitcode != 0:
        raise Exception('fail: {}'.format(out_to_str(stderr)))
    if tag is not None:
        tagged = out_to_str(stdout)[len(tag):]
        # If we're building directly from the tag we don't
        # need git-describe. In fact, PyPI won't like it
        # as it violates PEP 440. Let's return empty.
        if tagged[:3] == "-0-" and "dirty" not in tagged:
            return ""
        version = "+" + tagged[1:]
    else:
        version = "+g" + out_to_str(stdout)
    return version.replace('-', '.')


with open('LICENSE') as f:
    license = f.read()

with open('README.md') as f:
    readme = f.read()

version = {}
with open('adfotg/version.py', 'r') as f:
    exec(f.read(), version)


version_string = version['VERSION']
if os.path.exists(".git"):
    version_string += get_git_describe(get_git_tag())


setup(
    name=version['SHORTNAME'],
    version=version_string,
    description='ADF On-The-Go - convert your RPi Zero to Gotek-ready USB disk',
    long_description=readme,
    author='Robikz',
    author_email='zalewapl@gmail.com',
    license=license,
    include_package_data=True,
    packages=['adfotg'],
    url='https://github.com/Zalewa/adfotg',
    install_requires=[
        'Flask',
    ],
    entry_points={
        'console_scripts': ['adfotg=adfotg.cli:main']
    }
)



from . import error, version

from configparser import ConfigParser
import os
import sys

_PROGNAME = version.SHORTNAME
_CFG_FILENAME = "{}.conf".format(_PROGNAME)

DEFAULT_CONFIG_LOCATIONS = [os.path.join(dirp, _CFG_FILENAME) for dirp in [
    os.curdir,
    os.path.expanduser("~/.config"),
    "/etc/{}".format(_PROGNAME),
    "/etc"
]]


DEFAULT_PORT = 43164  # AMIGA

# TODO ${INSTALL_PREFIX}
_DATA_DIR = '/var/lib/{}'.format(_PROGNAME)
DEFAULT_ADF_DIR = os.path.join(_DATA_DIR, 'adf')
DEFAULT_UPLOAD_DIR = os.path.join(_DATA_DIR, 'upload')
DEFAULT_WORK_DIR = _DATA_DIR


class ConfigError(error.AdfotgError):
    pass


class _Config:
    def __init__(self):
        self.port = DEFAULT_PORT
        self.adf_dir = DEFAULT_ADF_DIR
        self.upload_dir = DEFAULT_UPLOAD_DIR
        self.work_dir = DEFAULT_WORK_DIR

    def load(self, parser):
        SECTION = _PROGNAME

        def _load(attr):
            setattr(self, attr, parser.get(
                SECTION, attr, fallback=getattr(self, attr)))

        self.port = parser.getint(SECTION, 'port', fallback=self.port)
        _load('adf_dir')
        _load('upload_dir')
        _load('work_dir')

    @property
    def mount_images_dir(self):
        return os.path.join(self.work_dir, "mount_images")

    @property
    def all_workspace_dirs(self):
        return [
            self.work_dir,
            self.adf_dir,
            self.upload_dir,
            self.mount_images_dir
        ]


def load(filenames=None):
    if not _load_from_file(filenames):
        _log("will attempt zeroconfig")
    _log_cfg()
    _create_env()


def _load_from_file(filenames=None):
    '''Config is loaded into the 'config' object which can be imported from
    this module elsewhere. Failure to load results in ConfigError throw.
    '''
    if filenames is None:
        locations = DEFAULT_CONFIG_LOCATIONS[:]
    else:
        if not isinstance(filenames, (list, tuple)):
            filenames = [filenames]
        locations = filenames
    filenames = [l for l in locations if os.path.isfile(l)]
    for filename in filenames:
        cfgparser = ConfigParser()
        if cfgparser.read(filename):
            config.load(cfgparser)
            _log("loaded configuration from '{}'".format(filename))
            return True
    _log("could not load config file from any of these "
         "locations: {}" .format(', '.join(locations)))
    return False


def _log_cfg():
    _log("Configuration:")
    for cfgvar in sorted(vars(config)):
        if cfgvar.startswith("_"):
            continue
        _log("  {} = {}".format(cfgvar, getattr(config, cfgvar)))


def _create_env():
    for dir_ in config.all_workspace_dirs:
        try:
            os.makedirs(dir_, exist_ok=True)
        except PermissionError as e:
            raise ConfigError(
                "could not create directory '{}'; ensure you are "
                "running as an appropriate user ({})".format(dir_, e)) from e


def _log(*args):
    print(*args, file=sys.stderr)


# This is a globally accessible object loaded at app startup.
config = _Config()



from . import app
from .error import ActionError

from flask import abort, jsonify, send_from_directory
from werkzeug.exceptions import NotFound

import os


@app.route('/', defaults={'path': 'index.html'})
@app.route('/<path:path>')
def serve_file(path):
    try:
        if os.path.isdir(os.path.join(app.root_path, "site")):
            # This is valid for deployment.
            return send_from_directory("site", path)
        elif os.path.isfile(os.path.join(app.root_path, "../setup.py")):
            # Assume development mode and try to send from repository.
            return send_from_directory("../site/dist", path)
        else:
            return abort(404)
    except NotFound as e:
        if path == "index.html":
            return abort(500, "index.html should have really been found."
                         " Your installation is broken")
        else:
            raise


@app.route('/adfwizard')
def adfwizard():
    return serve_file('index.html')


@app.errorhandler(500)
def error_500(exception):
    code = 500
    if isinstance(exception, ActionError):
        code = 400
    return (
        jsonify({"error": str(exception)}),
        code,
        {'Content-Type': 'application/json'}
    )
